#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import logging
import struct
from typing import Generator, Tuple, cast

from eth_typing import ChecksumAddress, HexStr
from hexbytes import HexBytes
from eth_account.datastructures import SignedMessage, SignedTransaction
from eth_account._utils.legacy_transactions import (
    encode_transaction,
    serializable_unsigned_transaction_from_dict as tx_from_dict,
    Transaction,
    UnsignedTransaction,
)
from eth_account.typed_transactions.typed_transaction import TypedTransaction

from eth_utils.crypto import keccak
from rlp import encode
from web3 import Web3
from web3.contract.contract import ContractFunction
from web3.exceptions import Web3Exception
from web3.types import _Hash32, TxParams, TxReceipt

import skale.config as config
from skale.transactions.exceptions import TransactionNotSentError, TransactionNotSignedError
from skale.utils.web3_utils import (
    DEFAULT_BLOCKS_TO_WAIT,
    MAX_WAITING_TIME,
    get_eth_nonce,
    public_key_to_address,
    to_checksum_address,
    wait_for_receipt_by_blocks,
)
from skale.wallets.common import BaseWallet, ensure_chain_id

logger = logging.getLogger(__name__)


class LedgerCommunicationError(Exception):
    pass


def encode_bip32_path(path: str) -> bytes:
    if len(path) == 0:
        return b''
    encoded_chunks = []
    for bip32_chunk in path.split('/'):
        chunk = bip32_chunk.split("'")
        if len(chunk) == 1:
            encoded_chunk = struct.pack('>I', int(chunk[0]))
        else:
            encoded_chunk = struct.pack('>I', 0x80000000 | int(chunk[0]))
        encoded_chunks.append(encoded_chunk)

    return b''.join(encoded_chunks)


def derivation_path_prefix(bin32_path: str) -> bytes:
    encoded_path = encode_bip32_path(bin32_path)
    encoded_path_len_bytes = (len(encoded_path) // 4).to_bytes(1, 'big')
    return encoded_path_len_bytes + encoded_path


def chunks(sequence: bytes, size: int) -> Generator[bytes, None, None]:
    return (sequence[pos : pos + size] for pos in range(0, len(sequence), size))


def get_derivation_path(address_index: int, legacy: bool) -> str:
    if legacy:
        return get_legacy_derivation_path(address_index)
    return get_live_derivation_path(address_index)


def get_live_derivation_path(address_index: int) -> str:
    return f"44'/60'/{address_index}'/0/0"


def get_legacy_derivation_path(address_index: int) -> str:
    return f"44'/60'/0'/{address_index}"


class LedgerWallet(BaseWallet):
    CHUNK_SIZE = 255
    CLA = b'\xe0'

    def __init__(self, web3: Web3, address_index: int, legacy: bool = False, debug: bool = False):
        from ledgerblue.comm import getDongle
        from ledgerblue.commException import CommException

        self._address_index = address_index
        self._bip32_path = get_derivation_path(address_index, legacy)
        try:
            self.dongle = getDongle(debug)
            self._web3 = web3
            self._address, self._public_key = self.get_address_with_public_key()
        except (OSError, CommException):
            raise LedgerCommunicationError(
                'Error occured during the interaction with Ledger device'
            )

    @property
    def address(self) -> ChecksumAddress:
        return self._address

    @property
    def public_key(self) -> str:
        return self._public_key

    # todo: remove this method after making software wallet as class
    def __getitem__(self, key: str) -> str:
        items = {'address': self.address, 'public_key': self.public_key}
        return items[key]

    def make_payload(self, data: str = '') -> bytes:
        encoded_data = cast(bytes, encode(data))
        path_prefix = derivation_path_prefix(self._bip32_path)
        return path_prefix + encoded_data

    @classmethod
    def parse_sign_result(
        cls,
        tx: TypedTransaction | Transaction | UnsignedTransaction,
        exchange_result: bytearray | bytes,
    ) -> SignedTransaction:
        sign_v = exchange_result[0]
        sign_r = int((exchange_result[1 : 1 + 32]).hex(), 16)
        sign_s = int((exchange_result[1 + 32 : 1 + 32 + 32]).hex(), 16)
        enctx = encode_transaction(tx, (sign_v, sign_r, sign_s))
        transaction_hash = keccak(enctx)

        return SignedTransaction(
            raw_transaction=HexBytes(enctx),
            hash=HexBytes(transaction_hash),
            v=sign_v,
            r=sign_r,
            s=sign_s,
        )

    def exchange_sign_payload_by_chunks(self, payload: bytes) -> bytearray:
        INS = b'\x04'
        P1_FIRST = b'\x00'
        P1_SUBSEQUENT = b'\x80'
        P2 = b'\x00'

        p1 = P1_FIRST
        for chunk in chunks(payload, LedgerWallet.CHUNK_SIZE):
            chunk_size_bytes = len(chunk).to_bytes(1, 'big')
            apdu = b''.join([LedgerWallet.CLA, INS, p1, P2, chunk_size_bytes, chunk])
            exchange_result = self.dongle.exchange(apdu)
            p1 = P1_SUBSEQUENT
        return cast(bytearray, exchange_result)

    def sign(self, tx_dict: TxParams) -> SignedTransaction:
        ensure_chain_id(tx_dict, self._web3)
        if tx_dict.get('nonce') is None:
            tx_dict['nonce'] = self._web3.eth.get_transaction_count(self.address)

        tx = tx_from_dict(tx_dict)
        try:
            payload = self.make_payload(tx)
            exchange_result = self.exchange_sign_payload_by_chunks(payload)
            return LedgerWallet.parse_sign_result(tx, exchange_result)
        except Exception as e:
            raise TransactionNotSignedError(e)

    def sign_and_send(
        self,
        tx: TxParams,
        multiplier: float | None = config.DEFAULT_GAS_MULTIPLIER,
        priority: int | None = config.DEFAULT_PRIORITY,
        method: str | None = None,
    ) -> HexStr:
        signed_tx = self.sign(tx)
        try:
            return Web3.to_hex(self._web3.eth.send_raw_transaction(signed_tx.raw_transaction))
        except (ValueError, Web3Exception) as e:
            raise TransactionNotSentError(e)

    def sign_hash(self, unsigned_hash: str) -> SignedMessage:
        raise NotImplementedError('sign_hash is not implemented for hardware wallet')

    @classmethod
    def parse_derive_result(cls, exchange_result: bytearray) -> Tuple[ChecksumAddress, str]:
        pk_len = exchange_result[0]
        pk = HexStr(exchange_result[1 : pk_len + 1].hex()[2:])
        address = public_key_to_address(pk)
        checksum_address = to_checksum_address(address)
        return checksum_address, pk

    def exchange_derive_payload(self, payload: bytes) -> bytearray:
        INS = b'\x02'
        P1 = b'\x00'
        P2 = b'\x00'
        payload_size_in_bytes = len(payload).to_bytes(1, 'big')
        apdu = b''.join([LedgerWallet.CLA, INS, P1, P2, payload_size_in_bytes, payload])
        return cast(bytearray, self.dongle.exchange(apdu))

    def get_address_with_public_key(self) -> tuple[ChecksumAddress, str]:
        payload = self.make_payload()
        exchange_result = self.exchange_derive_payload(payload)
        return LedgerWallet.parse_derive_result(exchange_result)

    def wait(
        self,
        tx_hash: _Hash32,
        blocks_to_wait: int = DEFAULT_BLOCKS_TO_WAIT,
        timeout: int = MAX_WAITING_TIME,
    ) -> TxReceipt:
        return wait_for_receipt_by_blocks(
            self._web3, tx_hash, blocks_to_wait=blocks_to_wait, timeout=timeout
        )


def hardware_sign_and_send(
    web3: Web3, method: ContractFunction, gas_amount: int, wallet: LedgerWallet
) -> str:
    address_from = wallet.address
    eth_nonce = get_eth_nonce(web3, address_from)
    tx_dict = method.build_transaction({'gas': gas_amount, 'nonce': eth_nonce})
    signed_txn = wallet.sign(tx_dict)
    tx = web3.eth.send_raw_transaction(signed_txn.raw_transaction).hex()
    logger.info(f'{method.__class__.__name__} - transaction_hash: {tx}')
    return tx



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
"""Account utilities"""

from __future__ import annotations
from decimal import Decimal
import logging
from typing import Any, Dict, List, Literal, Optional, TYPE_CHECKING, Type, TypedDict

from eth_typing import ChecksumAddress, HexStr
from web3 import Web3
from web3.types import TxReceipt, Wei

from skale.transactions.result import TxRes
from skale.transactions.tools import compose_eth_transfer_tx
from skale.utils.constants import LONG_LINE
from skale.wallets import LedgerWallet, Web3Wallet
from skale.utils.web3_utils import check_receipt, default_gas_price, wait_for_confirmation_blocks

if TYPE_CHECKING:
    from skale.skale_manager import SkaleManager
    from skale.wallets.common import BaseWallet


logger = logging.getLogger(__name__)


class AccountData(TypedDict):
    address: ChecksumAddress
    private_key: HexStr


WALLET_TYPE_TO_CLASS: Dict[str, Type[LedgerWallet] | Type[Web3Wallet]] = {
    'ledger': LedgerWallet,
    'web3': Web3Wallet,
}


def create_wallet(
    wallet_type: Literal['web3'] | Literal['ledger'] = 'web3', *args: Any, **kwargs: Any
) -> LedgerWallet | Web3Wallet:
    return WALLET_TYPE_TO_CLASS[wallet_type](*args, **kwargs)


def send_tokens(
    skale: SkaleManager, receiver_address: ChecksumAddress, amount: Wei, *args: Any, **kwargs: Any
) -> TxRes:
    logger.info(f'Sending {amount} SKALE tokens from {skale.wallet.address} => {receiver_address}')

    wei_amount = skale.web3.to_wei(amount, 'ether')
    return skale.token.transfer(receiver_address, wei_amount, *args, **kwargs)


def send_eth(
    web3: Web3,
    wallet: BaseWallet,
    receiver_address: ChecksumAddress,
    amount: Wei,
    *args: Any,
    gas_price: Optional[int] = None,
    wait_for: bool = True,
    confirmation_blocks: int = 0,
    multiplier: Optional[int] = None,
    priority: Optional[int] = None,
    **kwargs: Any,
) -> TxReceipt:
    logger.info(f'Sending {amount} ETH from {wallet.address} => {receiver_address}')
    wei_amount = web3.to_wei(amount, 'ether')
    gas_price = gas_price or default_gas_price(web3)
    tx = compose_eth_transfer_tx(
        web3, wallet.address, receiver_address, wei_amount, gas_price=gas_price, *args, **kwargs
    )
    tx_hash = wallet.sign_and_send(tx, multiplier=multiplier, priority=priority)
    if wait_for:
        receipt = wallet.wait(tx_hash)
    if confirmation_blocks:
        wait_for_confirmation_blocks(web3, confirmation_blocks)
    check_receipt(receipt)
    return receipt


def account_eth_balance_wei(web3: Web3, address: ChecksumAddress) -> Wei:
    return web3.eth.get_balance(address)


def check_ether_balance(web3: Web3, address: ChecksumAddress) -> int | Decimal:
    balance_wei = account_eth_balance_wei(web3, address)
    balance = web3.from_wei(balance_wei, 'ether')

    logger.info(f'{address} balance: {balance} ETH')
    return balance


def check_skale_balance(skale: SkaleManager, address: ChecksumAddress) -> int | Decimal:
    balance_wei = skale.token.get_balance(address)
    balance = skale.web3.from_wei(balance_wei, 'ether')
    logger.info(f'{address} balance: {balance} SKALE')
    return balance


def generate_account(web3: Web3) -> AccountData:
    account = web3.eth.account.create()
    private_key = account.key.hex()
    logger.info(f'Generated account: {account.address}')
    return AccountData({'address': account.address, 'private_key': private_key})


def generate_accounts(
    skale: SkaleManager,
    base_wallet: BaseWallet,
    n_wallets: int,
    skale_amount: Wei,
    eth_amount: Wei,
    debug: bool = False,
) -> List[AccountData]:
    n_wallets = int(n_wallets)
    results = []

    for _ in range(0, n_wallets):
        wallet = generate_account(skale.web3)

        send_tokens(skale, wallet['address'], skale_amount)
        send_eth(skale.web3, skale.wallet, wallet['address'], eth_amount)

        if debug:
            check_ether_balance(skale.web3, wallet['address'])
            check_skale_balance(skale, wallet['address'])

        results.append(wallet)
        logger.info(LONG_LINE)

    return results



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
"""SKALE Allocator Core Escrow methods"""

from typing import Any, Dict, List

from eth_typing import ChecksumAddress
from web3 import Web3
from web3.contract.contract import ContractFunction
from web3.exceptions import ContractLogicError
from web3.types import Wei

from skale.contracts.allocator_contract import AllocatorContract
from skale.contracts.base_contract import transaction_method
from skale.types.allocation import (
    BeneficiaryStatus,
    BeneficiaryPlan,
    Plan,
    PlanId,
    PlanWithId,
    TimeUnit,
)
from skale.utils.helper import format_fields


PLAN_FIELDS = [
    'totalVestingDuration',
    'vestingCliff',
    'vestingIntervalTimeUnit',
    'vestingInterval',
    'isDelegationAllowed',
    'isTerminatable',
]

BENEFICIARY_FIELDS = ['status', 'planId', 'startMonth', 'fullAmount', 'amountAfterLockup']

MAX_NUM_OF_PLANS = 9999
MAX_NUM_OF_BENEFICIARIES = 9999


class Allocator(AllocatorContract):
    def is_beneficiary_registered(self, beneficiary_address: ChecksumAddress) -> bool:
        """Confirms whether the beneficiary is registered in a Plan.

        :returns: Boolean value
        :rtype: bool
        """
        return bool(self.contract.functions.isBeneficiaryRegistered(beneficiary_address).call())

    def is_delegation_allowed(self, beneficiary_address: ChecksumAddress) -> bool:
        return bool(self.contract.functions.isDelegationAllowed(beneficiary_address).call())

    def is_vesting_active(self, beneficiary_address: ChecksumAddress) -> bool:
        return bool(self.contract.functions.isVestingActive(beneficiary_address).call())

    def get_escrow_address(self, beneficiary_address: ChecksumAddress) -> ChecksumAddress:
        return Web3.to_checksum_address(
            self.contract.functions.getEscrowAddress(beneficiary_address).call()
        )

    @transaction_method
    def add_plan(
        self,
        vesting_cliff: int,
        total_vesting_duration: int,
        vesting_interval_time_unit: TimeUnit,
        vesting_interval: int,
        can_delegate: bool,
        is_terminatable: bool,
    ) -> 'ContractFunction':
        return self.contract.functions.addPlan(
            vestingCliff=vesting_cliff,
            totalVestingDuration=total_vesting_duration,
            vestingIntervalTimeUnit=vesting_interval_time_unit.value,
            vestingInterval=vesting_interval,
            canDelegate=can_delegate,
            isTerminatable=is_terminatable,
        )

    @transaction_method
    def connect_beneficiary_to_plan(
        self,
        beneficiary_address: ChecksumAddress,
        plan_id: int,
        start_month: int,
        full_amount: int,
        lockup_amount: int,
    ) -> 'ContractFunction':
        return self.contract.functions.connectBeneficiaryToPlan(
            beneficiary=beneficiary_address,
            planId=plan_id,
            startMonth=start_month,
            fullAmount=full_amount,
            lockupAmount=lockup_amount,
        )

    @transaction_method
    def start_vesting(self, beneficiary_address: ChecksumAddress) -> 'ContractFunction':
        return self.contract.functions.startVesting(beneficiary_address)

    @transaction_method
    def stop_vesting(self, beneficiary_address: ChecksumAddress) -> 'ContractFunction':
        return self.contract.functions.stopVesting(beneficiary_address)

    @transaction_method
    def grant_role(self, role: bytes, address: ChecksumAddress) -> 'ContractFunction':
        return self.contract.functions.grantRole(role, address)

    def vesting_manager_role(self) -> bytes:
        return bytes(self.contract.functions.VESTING_MANAGER_ROLE().call())

    def has_role(self, role: bytes, address: ChecksumAddress) -> bool:
        return bool(self.contract.functions.hasRole(role, address).call())

    def __get_beneficiary_plan_params_raw(self, beneficiary_address: ChecksumAddress) -> List[Any]:
        return list(self.contract.functions.getBeneficiaryPlanParams(beneficiary_address).call())

    @format_fields(BENEFICIARY_FIELDS)
    def get_beneficiary_plan_params_dict(self, beneficiary_address: ChecksumAddress) -> List[Any]:
        return self.__get_beneficiary_plan_params_raw(beneficiary_address)

    def get_beneficiary_plan_params(self, beneficiary_address: ChecksumAddress) -> BeneficiaryPlan:
        plan_params = self.get_beneficiary_plan_params_dict(beneficiary_address)
        if plan_params is None:
            raise ValueError('Plan for ', beneficiary_address, ' is missing')
        if isinstance(plan_params, list):
            return self._to_beneficiary_plan(
                {**plan_params[0], 'statusName': BeneficiaryStatus(plan_params[0]['status']).name}
            )
        if isinstance(plan_params, dict):
            return self._to_beneficiary_plan(
                {**plan_params, 'statusName': BeneficiaryStatus(plan_params.get('status', 0)).name}
            )
        raise TypeError(f'Internal error on getting plan params for ${beneficiary_address}')

    def __get_plan_raw(self, plan_id: PlanId) -> List[Any]:
        return list(self.contract.functions.getPlan(plan_id).call())

    @format_fields(PLAN_FIELDS)
    def get_untyped_plan(self, plan_id: PlanId) -> List[Any]:
        return self.__get_plan_raw(plan_id)

    def get_plan(self, plan_id: PlanId) -> Plan:
        untyped_plan = self.get_untyped_plan(plan_id)
        if untyped_plan is None:
            raise ValueError('Plan ', plan_id, ' is missing')
        if isinstance(untyped_plan, list):
            return self._to_plan(untyped_plan[0])
        if isinstance(untyped_plan, dict):
            return self._to_plan(untyped_plan)
        raise TypeError(plan_id)

    def get_all_plans(self) -> List[PlanWithId]:
        plans = []
        for i in range(1, MAX_NUM_OF_PLANS):
            try:
                plan_id = PlanId(i)
                plan = PlanWithId({**self.get_plan(plan_id), 'planId': plan_id})
                plans.append(plan)
            except (ContractLogicError, ValueError):
                break
        return plans

    def calculate_vested_amount(self, address: ChecksumAddress) -> Wei:
        return Wei(self.contract.functions.calculateVestedAmount(address).call())

    def get_finish_vesting_time(self, address: ChecksumAddress) -> int:
        return int(self.contract.functions.getFinishVestingTime(address).call())

    def get_lockup_period_end_timestamp(self, address: ChecksumAddress) -> int:
        return int(self.contract.functions.getLockupPeriodEndTimestamp(address).call())

    def get_time_of_next_vest(self, address: ChecksumAddress) -> int:
        return int(self.contract.functions.getTimeOfNextVest(address).call())

    def _to_plan(self, untyped_plan: Dict[str, Any]) -> Plan:
        return Plan(
            {
                'totalVestingDuration': int(untyped_plan['totalVestingDuration']),
                'vestingCliff': int(untyped_plan['vestingCliff']),
                'vestingIntervalTimeUnit': TimeUnit(untyped_plan['vestingIntervalTimeUnit']),
                'vestingInterval': int(untyped_plan['vestingInterval']),
                'isDelegationAllowed': bool(untyped_plan['isDelegationAllowed']),
                'isTerminatable': bool(untyped_plan['isTerminatable']),
            }
        )

    def _to_beneficiary_plan(self, untyped_beneficiary_plan: Dict[str, Any]) -> BeneficiaryPlan:
        return BeneficiaryPlan(
            {
                'status': BeneficiaryStatus(untyped_beneficiary_plan['status']),
                'statusName': str(untyped_beneficiary_plan['statusName']),
                'planId': PlanId(untyped_beneficiary_plan['planId']),
                'startMonth': int(untyped_beneficiary_plan['startMonth']),
                'fullAmount': Wei(untyped_beneficiary_plan['fullAmount']),
                'amountAfterLockup': Wei(untyped_beneficiary_plan['amountAfterLockup']),
            }
        )



from web3.exceptions import ContractLogicError


class TransactionError(Exception):
    """
    Base exception for transaction related errors
    """

    pass


class ChainIdError(TransactionError):
    """
    Raised when chainId is missing or incorrect
    """

    pass


class TransactionNotSignedError(TransactionError):
    """
    Raised when transaction wasn't signed
    """

    pass


class TransactionNotSentError(TransactionError):
    """
    Raised when transaction wasn't sent
    """

    pass


class TransactionNotMinedError(TimeoutError, TransactionError):
    """
    Raised when transaction wasn't included in block within timeout
    """

    pass


class TransactionWaitError(TimeoutError, TransactionError):
    """
    Raised when error occurred during waiting for transaction
    to be included in block
    """

    pass


class TransactionLogicError(TransactionError, ContractLogicError):
    """
    Raised when transaction executed with error
    """

    pass


class DryRunFailedError(TransactionLogicError):
    """
    Raised when error occurred during dry run call
    """

    pass


class TransactionFailedError(TransactionLogicError):
    """
    Raised when transaction included in the block failed during execution
    """

    pass


class TransactionRevertError(TransactionFailedError):
    """
    Raised when transaction included in the block failed with revert
    """

    pass


class DryRunRevertError(DryRunFailedError):
    """
    Raised when transaction reverted during dry run call
    """

    pass



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2021 SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from skale.types.schain import SchainOption
from enum import Enum


class AllocationType(int, Enum):
    DEFAULT = 0
    NO_FILESTORAGE = 1
    MAX_CONTRACT_STORAGE = 2
    MAX_CONSENSUS_DB = 3
    MAX_FILESTORAGE = 4


@dataclass
class SchainOptions:
    multitransaction_mode: bool
    threshold_encryption: bool
    allocation_type: AllocationType

    def to_tuples(self) -> list[SchainOption]:
        return [
            ('multitr', bool_to_bytes(self.multitransaction_mode)),
            ('encrypt', bool_to_bytes(self.threshold_encryption)),
            ('alloc', int_to_bytes(self.allocation_type.value)),
        ]


def parse_schain_options(raw_options: list[SchainOption]) -> SchainOptions:
    """
    Parses raw sChain options from smart contracts (list of tuples).
    Returns default values if nothing is set on contracts.
    """

    multitransaction_mode = False
    threshold_encryption = False
    allocation_type = AllocationType.DEFAULT
    if len(raw_options) > 0:
        multitransaction_mode = bytes_to_bool(raw_options[0][1])
    if len(raw_options) > 1:
        threshold_encryption = bytes_to_bool(raw_options[1][1])
    if len(raw_options) > 2:
        allocation_type = AllocationType(bytes_to_int(raw_options[2][1]))

    return SchainOptions(
        multitransaction_mode=multitransaction_mode,
        threshold_encryption=threshold_encryption,
        allocation_type=allocation_type,
    )


def get_default_schain_options() -> SchainOptions:
    return SchainOptions(
        multitransaction_mode=False,
        threshold_encryption=False,
        allocation_type=AllocationType.DEFAULT,
    )


def bool_to_bytes(bool_value: bool) -> bytes:
    return bool_value.to_bytes(1, byteorder='big')


def int_to_bytes(int_value: int) -> bytes:
    return int.to_bytes(int_value, length=1, byteorder='big')


def bytes_to_int(bytes_value: bytes) -> int:
    return int.from_bytes(bytes_value, byteorder='big')


def bytes_to_bool(bytes_value: bytes) -> bool:
    return bool(int.from_bytes(bytes_value, 'big'))



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import random
from skale.utils.random_names.vocabulary import (
    CONSTELLATIONS,
    STARS,
    SOUND_AND_APPEARANCE_ADJECTIVES,
    POSITIVE_AND_TIME_ADJECTIVES,
)


def generate_random_node_name() -> str:
    return generate_name(CONSTELLATIONS, POSITIVE_AND_TIME_ADJECTIVES)


def generate_random_schain_name() -> str:
    return generate_name(STARS, SOUND_AND_APPEARANCE_ADJECTIVES)


def generate_name(noun_dict: list[str], adjective_dict: list[str]) -> str:
    noun = get_random_word_from_dict(noun_dict)
    adjective = get_random_word_from_dict(adjective_dict)
    return f'{adjective}-{noun}'


def get_random_word_from_dict(vocabulary_dict: list[str]) -> str:
    return random.choice(vocabulary_dict).lower().replace(' ', '-')


if __name__ == '__main__':
    node_name = generate_random_node_name()
    schain_name = generate_random_schain_name()
    print(f'Node name: {node_name}, sChain name: {schain_name}')



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from skale.contracts.base_contract import BaseContract, transaction_method
from eth_typing import ChecksumAddress
from skale.transactions.result import TxRes


class ConfigController(BaseContract):
    """Config controller contract"""

    def default_admin_role(self) -> bytes:
        return self.contract.functions.DEFAULT_ADMIN_ROLE().call()

    def deployer_admin_role(self) -> bytes:
        return self.contract.functions.DEPLOYER_ADMIN_ROLE().call()

    def deployer_role(self) -> bytes:
        return self.contract.functions.DEPLOYER_ROLE().call()

    def mtm_admin_role(self) -> bytes:
        return self.contract.functions.MTM_ADMIN_ROLE().call()

    def allowed_origin_role(self, deployer: ChecksumAddress) -> bytes:
        return self.contract.functions.allowedOriginRole(deployer).call()

    def allowed_origin_role_admin(self, deployer: ChecksumAddress) -> bytes:
        return self.contract.functions.allowedOriginRoleAdmin(deployer).call()

    def has_role(self, role: bytes, address: ChecksumAddress) -> bool:
        return bool(self.contract.functions.hasRole(role, address).call())

    def get_role_admin(self, role: bytes) -> bytes:
        return self.contract.functions.getRoleAdmin(role).call()

    def get_role_member(self, role: bytes, index: int) -> bytes:
        return self.contract.functions.getRoleMember(role, index).call()

    def get_role_member_count(self, role: bytes) -> int:
        return self.contract.functions.getRoleMemberCount(role).call()

    def is_address_whitelisted(self, address: ChecksumAddress) -> bool:
        return bool(self.contract.functions.isAddressWhitelisted(address).call())

    def is_deployment_allowed(
        self, transaction_origin: ChecksumAddress, deployer: ChecksumAddress
    ) -> bool:
        return bool(
            self.contract.functions.isDeploymentAllowed(transaction_origin, deployer).call()
        )

    @transaction_method
    def grant_role(self, role: bytes, address: ChecksumAddress) -> TxRes:
        return self.contract.functions.grantRole(role, address)

    @transaction_method
    def add_allowed_origin_role_admin(self, role: bytes, address: ChecksumAddress) -> TxRes:
        return self.contract.functions.addAllowedOriginRoleAdmin(role, address)

    @transaction_method
    def allow_origin(self, transaction_origin: ChecksumAddress, deployer: ChecksumAddress) -> TxRes:
        return self.contract.functions.allowOrigin(transaction_origin, deployer)

    @transaction_method
    def add_to_whitelist(self, address: ChecksumAddress) -> TxRes:
        return self.contract.functions.addToWhitelist(address)

    @transaction_method
    def revoke_role(self, role: bytes, address: ChecksumAddress) -> TxRes:
        return self.contract.functions.revokeRole(role, address)

    @transaction_method
    def renounce_role(self, role: bytes, address: ChecksumAddress) -> TxRes:
        return self.contract.functions.renounceRole(role, address)

    @transaction_method
    def remove_allowed_origin_role_admin(self, role: bytes, address: ChecksumAddress) -> TxRes:
        return self.contract.functions.removeAllowedOriginRoleAdmin(role, address)

    @transaction_method
    def forbid_origin(
        self, transaction_origin: ChecksumAddress, deployer: ChecksumAddress
    ) -> TxRes:
        return self.contract.functions.forbidOrigin(transaction_origin, deployer)

    @transaction_method
    def remove_from_whitelist(self, address: ChecksumAddress) -> TxRes:
        return self.contract.functions.removeFromWhitelist(address)

    @transaction_method
    def enable_free_contract_deployment(self) -> TxRes:
        return self.contract.functions.enableFreeContractDeployment()

    @transaction_method
    def disable_free_contract_deployment(self) -> TxRes:
        return self.contract.functions.disableFreeContractDeployment()

    def is_fcd_enabled(self) -> str:
        return self.contract.functions.isFCDEnabled().call()

    @transaction_method
    def enable_mtm(self) -> TxRes:
        return self.contract.functions.enableMTM()

    @transaction_method
    def disable_mtm(self) -> TxRes:
        return self.contract.functions.disableMTM()

    def is_mtm_enabled(self) -> str:
        return self.contract.functions.isMTMEnabled().call()

    def get_version(self) -> str:
        return self.contract.functions.version().call()

    @transaction_method
    def set_version(self, new_version) -> TxRes:
        return self.contract.functions.setVersion(new_version)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2025-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import socket
from typing import Any, List

from eth_typing import ChecksumAddress, HexStr
from skale.contracts.base_contract import BaseContract
from skale.types.node import MirageNode, NodeId, Port
from skale.contracts.base_contract import transaction_method
from skale.utils import helper


class Nodes(BaseContract):
    def __get_raw(self, node_id: NodeId) -> List[Any]:
        return list(self.contract.functions.getNode(node_id).call())

    def get(self, node_id: NodeId) -> MirageNode:
        return self._to_node(self.__get_raw(node_id))

    def get_id(self, address: ChecksumAddress) -> NodeId:
        return self.contract.functions.getNodeId(address).call()

    def get_by_address(self, address: ChecksumAddress) -> MirageNode:
        node_id = self.get_id(address)
        return self.get(node_id)

    def get_passive_node_ids_for_address(self, node_address: ChecksumAddress) -> list[NodeId]:
        return self.contract.functions.getPassiveNodesIdsForAddress(node_address).call()

    def get_passive_node_ids(self) -> list[NodeId]:
        return self.contract.functions.getPassiveNodesIds().call()

    def get_active_node_ids(self) -> list[NodeId]:
        return self.contract.functions.getActiveNodesIds().call()

    def active_node_exists(self, node_id: NodeId) -> bool:
        return self.contract.functions.activeNodeExists(node_id).call()

    def _to_node(self, untyped_node: List[Any]) -> MirageNode:
        return MirageNode(
            id=untyped_node[0],
            public_key=self.decode_public_key(untyped_node[1]),
            ip=bytes(untyped_node[2]),
            ip_str=socket.inet_ntoa(untyped_node[2]),
            domain_name=untyped_node[3],
            address=ChecksumAddress(untyped_node[4]),
            port=Port(untyped_node[5]),
            name=f'node-{untyped_node[0]}',
        )

    def decode_public_key(self, raw_public_key: list[bytes]) -> HexStr:
        key_bytes = raw_public_key[0] + raw_public_key[1]
        return self.skale.web3.to_hex(key_bytes)

    @transaction_method
    def register_active(self, ip: str, port: Port):
        ip_bytes = socket.inet_aton(ip)
        pk_parts_bytes = helper.split_public_key(self.skale.wallet.public_key)
        return self.contract.functions.registerNode(ip_bytes, pk_parts_bytes, port)

    @transaction_method
    def register_passive(self, ip: str, port: Port):
        ip_bytes = socket.inet_aton(ip)
        return self.contract.functions.registerPassiveNode(ip_bytes, port)

    @transaction_method
    def request_change_owner(self, node_id: NodeId, new_owner: ChecksumAddress):
        return self.contract.functions.requestChangeOwner(node_id, new_owner)

    @transaction_method
    def confirm_owner_change(self, node_id: NodeId):
        return self.contract.functions.confirmOwnerChange(node_id)

    @transaction_method
    def set_domain_name(self, node_id: NodeId, domain_name: str):
        return self.contract.functions.setDomainName(node_id, domain_name)

    @transaction_method
    def set_ip_address(self, node_id: NodeId, ip: str, port: Port):
        ip_bytes = socket.inet_aton(ip)
        return self.contract.functions.setIpAddress(node_id, ip_bytes, port)

    @transaction_method
    def set_committee(self, committee_address: ChecksumAddress):
        return self.contract.functions.setCommittee(committee_address)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from eth_typing import ChecksumAddress
from skale.contracts.base_contract import BaseContract, transaction_method
from skale.transactions.result import TxRes
from skale.types.schain import SchainHash, SchainName
from Crypto.Hash import keccak


class Paymaster(BaseContract):
    """Paymaster contract"""

    def name_to_id(self, schain_name: SchainName) -> SchainHash:
        keccak_hash = keccak.new(data=schain_name.encode('utf8'), digest_bits=256)
        return keccak_hash.digest()

    @transaction_method
    def add_schain(self, schain_name: SchainName) -> TxRes:
        return self.contract.functions.addSchain(schain_name)

    @transaction_method
    def remove_schain(self, schain_name: SchainName) -> TxRes:
        schain_id = self.name_to_id(schain_name)
        return self.contract.functions.removeSchain(schain_id)

    @transaction_method
    def add_validator(self, validator_id: int, validator_address: ChecksumAddress) -> TxRes:
        return self.contract.functions.addValidator(validator_id, validator_address)

    @transaction_method
    def remove_validator(self, validator_id: int) -> TxRes:
        return self.contract.functions.removeValidator(validator_id)

    @transaction_method
    def set_validator_address(self, validator_id: int, new_address: ChecksumAddress) -> TxRes:
        return self.contract.functions.setValidatorAddress(validator_id, new_address)

    @transaction_method
    def set_active_nodes(self, validator_id: int, nodes_amount: int) -> TxRes:
        return self.contract.functions.setActiveNodes(validator_id, nodes_amount)

    @transaction_method
    def set_max_replenishment_period(self, month: int) -> TxRes:
        return self.contract.functions.setMaxReplenishmentPeriod(month)

    @transaction_method
    def set_schain_price(self, price: int) -> TxRes:
        return self.contract.functions.setSchainPrice(price)

    @transaction_method
    def set_skl_price(self, price: int) -> TxRes:
        return self.contract.functions.setSklPrice(price)

    @transaction_method
    def set_allowed_skl_price_lag(self, lag_seconds: int) -> TxRes:
        return self.contract.functions.setAllowedSklPriceLag(lag_seconds)

    @transaction_method
    def set_skale_token(self, token_address: ChecksumAddress) -> TxRes:
        return self.contract.functions.setSkaleToken(token_address)

    @transaction_method
    def clear_history(self, timestamp_before: int) -> TxRes:
        return self.contract.functions.clearHistory(timestamp_before)

    @transaction_method
    def pay(self, schain_name: SchainName, month: int) -> TxRes:
        schain_id = self.name_to_id(schain_name)
        return self.contract.functions.pay(schain_id, month)

    @transaction_method
    def claim(self, to_address: ChecksumAddress) -> TxRes:
        return self.contract.functions.claim(to_address)

    @transaction_method
    def set_version(self, new_version) -> TxRes:
        return self.contract.functions.setVersion(new_version)

    def get_schain_expiration_timestamp(self, schain_name: SchainName) -> TxRes:
        schain_id = self.name_to_id(schain_name)
        return self.contract.functions.getSchainExpirationTimestamp(schain_id).call()

    def get_reward_amount(self, validator_id: int) -> int:
        return self.contract.functions.getRewardAmount(validator_id).call()

    def get_nodes_number(self, validator_id: int) -> int:
        return self.contract.functions.getNodesNumber(validator_id).call()

    def get_historical_active_nodes_number(self, validator_id: int, time_when: int) -> int:
        return self.contract.functions.getHistoricalActiveNodesNumber(
            validator_id, time_when
        ).call()

    def get_historical_total_active_nodes_number(self, time_when: int) -> int:
        return self.contract.functions.getHistoricalTotalActiveNodesNumber(time_when).call()

    def get_validator_number(self) -> int:
        return self.contract.functions.getValidatorsNumber().call()

    def get_schain_names(self) -> str:
        return self.contract.functions.getSchainsNames().call()

    def get_total_reward(self, time_from: int, time_to: int) -> int:
        return self.contract.functions.getTotalReward(time_from, time_to).call()

    @transaction_method
    def set_nodes_amount(self, validator_id: int, amount: int) -> TxRes:
        return self.contract.functions.setNodesAmount(validator_id, amount)

    @transaction_method
    def claim_for(self, validator_id: int, address: ChecksumAddress) -> TxRes:
        return self.contract.functions.claimFor(validator_id, address)

    def get_schain_number(self) -> str:
        return self.contract.functions.getSchainsNumber().call()

    def get_schain(self, schain_name: SchainName) -> list:
        schain_id = self.name_to_id(schain_name)
        return self.contract.functions.schains(schain_id).call()

    def get_max_replenishment_period(self) -> int:
        return self.contract.functions.maxReplenishmentPeriod().call()

    def get_schain_price_per_month(self) -> int:
        return self.contract.functions.schainPricePerMonth().call()

    def get_one_skl_price(self) -> int:
        return self.contract.functions.oneSklPrice().call()

    def get_skl_price_timestamp(self) -> int:
        return self.contract.functions.sklPriceTimestamp().call()

    def get_allowed_skl_price_lag(self) -> int:
        return self.contract.functions.allowedSklPriceLag().call()

    def get_skale_token(self) -> int:
        return self.contract.functions.skaleToken().call()

    @transaction_method
    def skip_time(self, seconds: int) -> TxRes:
        return self.contract.functions.skipTime(seconds)

    def get_effective_timestamp(self) -> int:
        return self.contract.functions.effectiveTimestamp().call()

    def get_debts_amount(self, debt_id: int) -> int:
        return self.contract.functions.debts(debt_id).call()

    def get_debts_begin(self) -> int:
        return self.contract.functions.debtsBegin().call()

    def get_debts_end(self) -> int:
        return self.contract.functions.debtsEnd().call()



import sys

if sys.version_info < (3, 7):
    raise EnvironmentError('Python 3.7 or above is required')

from skale.skale_manager import SkaleManager
from skale.skale_allocator import SkaleAllocator
from skale.skale_ima import SkaleIma
from skale.mirage_manager import MirageManager

__all__ = ['SkaleManager', 'SkaleAllocator', 'SkaleIma', 'MirageManager']



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2025-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from skale.contracts.base_contract import BaseContract
from skale.contracts.base_contract import transaction_method
from skale.types.node import NodeId


class Status(BaseContract):
    def is_healthy(self, node_id: NodeId) -> bool:
        return self.contract.functions.isHealthy(node_id).call()

    def get_nodes_eligible_for_committee(self) -> list[NodeId]:
        return self.contract.functions.getNodesEligibleForCommittee().call()

    def get_whitelisted_nodes(self) -> list[NodeId]:
        return self.contract.functions.getWhitelistedNodes().call()

    def is_whitelisted(self, node_id: NodeId) -> bool:
        return self.contract.functions.isWhitelisted(node_id).call()

    @transaction_method
    def alive(self):
        return self.contract.functions.alive()

    @transaction_method
    def set_heartbeat_interval(self, interval: int):
        return self.contract.functions.setHeartbeatInterval(interval)

    @transaction_method
    def whitelist_node(self, node_id: NodeId):
        return self.contract.functions.whitelistNode(node_id)

    @transaction_method
    def remove_node_from_whitelist(self, node_id: NodeId):
        return self.contract.functions.removeNodeFromWhitelist(node_id)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from typing import List, Tuple
from eth_typing import ChecksumAddress
from web3.contract.contract import ContractFunction

from skale.contracts.base_contract import transaction_method
from skale.contracts.skale_manager_contract import SkaleManagerContract
from skale.transactions.tools import retry_tx
from skale.types.dkg import G2Point, KeyShare, VerificationVector
from skale.types.node import NodeId
from skale.types.schain import SchainHash


class DKG(SkaleManagerContract):
    @retry_tx
    @transaction_method
    def broadcast(
        self,
        group_index: SchainHash,
        node_index: NodeId,
        verification_vector: VerificationVector,
        secret_key_contribution: List[KeyShare],
        rotation_id: int,
    ) -> 'ContractFunction':
        return self.contract.functions.broadcast(
            group_index, node_index, verification_vector, secret_key_contribution, rotation_id
        )

    @retry_tx
    @transaction_method
    def pre_response(
        self,
        group_index: SchainHash,
        from_node_index: NodeId,
        verification_vector: VerificationVector,
        verification_vector_mult: VerificationVector,
        secret_key_contribution: List[KeyShare],
    ) -> 'ContractFunction':
        return self.contract.functions.preResponse(
            group_index,
            fromNodeIndex=from_node_index,
            verificationVector=verification_vector,
            verificationVectorMultiplication=verification_vector_mult,
            secretKeyContribution=secret_key_contribution,
        )

    @retry_tx
    @transaction_method
    def response(
        self,
        group_index: SchainHash,
        from_node_index: NodeId,
        secret_number: int,
        multiplied_share: G2Point,
    ) -> 'ContractFunction':
        return self.contract.functions.response(
            group_index,
            fromNodeIndex=from_node_index,
            secretNumber=secret_number,
            multipliedShare=multiplied_share,
        )

    @retry_tx
    @transaction_method
    def alright(self, group_index: SchainHash, from_node_index: NodeId) -> 'ContractFunction':
        return self.contract.functions.alright(group_index, from_node_index)

    @retry_tx
    @transaction_method
    def complaint(
        self, group_index: SchainHash, from_node_index: NodeId, to_node_index: NodeId
    ) -> 'ContractFunction':
        return self.contract.functions.complaint(group_index, from_node_index, to_node_index)

    @retry_tx
    @transaction_method
    def complaint_bad_data(
        self, group_index: SchainHash, from_node_index: NodeId, to_node_index: NodeId
    ) -> 'ContractFunction':
        return self.contract.functions.complaintBadData(group_index, from_node_index, to_node_index)

    def is_last_dkg_successful(self, group_index: SchainHash) -> bool:
        return bool(self.contract.functions.isLastDKGSuccessful(group_index).call())

    def is_channel_opened(self, group_index: SchainHash) -> bool:
        return bool(self.contract.functions.isChannelOpened(group_index).call())

    def is_broadcast_possible(
        self, group_index: SchainHash, node_id: NodeId, address: ChecksumAddress
    ) -> bool:
        return bool(
            self.contract.functions.isBroadcastPossible(group_index, node_id).call(
                {'from': address}
            )
        )

    def is_alright_possible(
        self, group_index: SchainHash, node_id: NodeId, address: ChecksumAddress
    ) -> bool:
        return bool(
            self.contract.functions.isAlrightPossible(group_index, node_id).call({'from': address})
        )

    def is_complaint_possible(
        self, group_index: SchainHash, node_from: NodeId, node_to: NodeId, address: ChecksumAddress
    ) -> bool:
        return bool(
            self.contract.functions.isComplaintPossible(group_index, node_from, node_to).call(
                {'from': address}
            )
        )

    def is_pre_response_possible(
        self, group_index: SchainHash, node_id: NodeId, address: ChecksumAddress
    ) -> bool:
        return bool(
            self.contract.functions.isPreResponsePossible(group_index, node_id).call(
                {'from': address}
            )
        )

    def is_response_possible(
        self, group_index: SchainHash, node_id: NodeId, address: ChecksumAddress
    ) -> bool:
        return bool(
            self.contract.functions.isResponsePossible(group_index, node_id).call({'from': address})
        )

    def is_all_data_received(self, group_index: SchainHash, node_from: NodeId) -> bool:
        return bool(self.contract.functions.isAllDataReceived(group_index, node_from).call())

    def is_everyone_broadcasted(self, group_index: SchainHash, address: ChecksumAddress) -> bool:
        return bool(
            self.contract.functions.isEveryoneBroadcasted(group_index).call({'from': address})
        )

    def get_number_of_completed(self, group_index: SchainHash) -> int:
        return int(self.contract.functions.getNumberOfCompleted(group_index).call())

    def get_channel_started_time(self, group_index: SchainHash) -> int:
        return int(self.contract.functions.getChannelStartedTime(group_index).call())

    def get_complaint_started_time(self, group_index: SchainHash) -> int:
        return int(self.contract.functions.getComplaintStartedTime(group_index).call())

    def get_alright_started_time(self, group_index: SchainHash) -> int:
        return int(self.contract.functions.getAlrightStartedTime(group_index).call())

    def get_complaint_data(self, group_index: SchainHash) -> Tuple[NodeId, NodeId]:
        return tuple(self.contract.functions.getComplaintData(group_index).call())

    def get_time_of_last_successful_dkg(self, group_index: SchainHash) -> int:
        return int(self.contract.functions.getTimeOfLastSuccessfulDKG(group_index).call())

    def is_node_broadcasted(self, group_index: SchainHash, node_id: NodeId) -> bool:
        return bool(self.contract.functions.isNodeBroadcasted(group_index, node_id).call())



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2024-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from collections import namedtuple
from dataclasses import dataclass
from typing import TypedDict

from skale.types.node import NodeId
from skale.types.schain import SchainHash


RotationNodeData = namedtuple('RotationNodeData', ['index', 'node_id', 'public_key'])


class NodesSwap(TypedDict):
    leaving_node_id: NodeId
    new_node_id: NodeId


class BlsPublicKey(TypedDict):
    blsPublicKey0: str
    blsPublicKey1: str
    blsPublicKey2: str
    blsPublicKey3: str


class NodesGroup(TypedDict):
    rotation: NodesSwap | None
    nodes: dict[NodeId, RotationNodeData]
    finish_ts: int | None
    bls_public_key: BlsPublicKey | None


@dataclass
class Rotation:
    leaving_node_id: NodeId
    new_node_id: NodeId
    freeze_until: int
    rotation_counter: int


class RotationSwap(TypedDict):
    schain_id: SchainHash
    finished_rotation: int



from skale.contracts.manager.contract_manager import ContractManager
from skale.contracts.base_contract import BaseContract, transaction_method

from skale.contracts.ima.mainnet.linker import Linker

__all__ = ['Linker', 'ContractManager', 'BaseContract', 'transaction_method']



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from functools import cached_property

from skale_contracts.projects.config_controller import ConfigControllerContract
from skale_contracts.project_factory import SkaleProject

from skale.skale_base import SkaleBase
from skale.contracts.config_controller.config_controller import ConfigController


class SkaleConfigController(SkaleBase):
    @property
    def project_name(self) -> SkaleProject:
        return SkaleProject.CONFIG_CONTROLLER

    @cached_property
    def config_controller(self) -> ConfigController:
        return ConfigController(self, ConfigControllerContract.CONFIG_CONTROLLER)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.


from collections import namedtuple
from typing import List
from web3.contract.contract import ContractFunction

from skale.contracts.base_contract import transaction_method
from skale.contracts.skale_manager_contract import SkaleManagerContract
from skale.transactions.result import TxRes
from skale.utils.helper import ip_from_bytes, ip_to_bytes
from eth_typing import ChecksumAddress


class IpRange(namedtuple('IpRange', ['start_ip', 'end_ip'])):
    @classmethod
    def from_packed(cls, packed_ips: List[bytes]) -> 'IpRange':
        return cls(ip_from_bytes(packed_ips[0]), ip_from_bytes(packed_ips[1]))


class SyncManager(SkaleManagerContract):
    """Wrapper for SyncManager.sol functions"""

    @transaction_method
    def add_ip_range(self, name: str, start_ip: str, end_ip: str) -> 'ContractFunction':
        return self.contract.functions.addIPRange(name, ip_to_bytes(start_ip), ip_to_bytes(end_ip))

    @transaction_method
    def remove_ip_range(self, name: str) -> 'ContractFunction':
        return self.contract.functions.removeIPRange(name)

    def get_ip_ranges_number(self) -> int:
        return int(self.contract.functions.getIPRangesNumber().call())

    def get_ip_range_by_index(self, index: int) -> IpRange:
        packed = self.contract.functions.getIPRangeByIndex(index).call()
        return IpRange.from_packed(packed)

    def get_ip_range_by_name(self, name: str) -> IpRange:
        packed = self.contract.functions.getIPRangeByName(name).call()
        return IpRange.from_packed(packed)

    def grant_sync_manager_role(self, address: ChecksumAddress) -> TxRes:
        return self.grant_role(self.sync_manager_role(), address)

    def sync_manager_role(self) -> bytes:
        return bytes(self.contract.functions.SYNC_MANAGER_ROLE().call())

    @transaction_method
    def grant_role(self, role: bytes, owner: str) -> 'ContractFunction':
        return self.contract.functions.grantRole(role, owner)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
"""Schains.sol functions"""

import functools
from dataclasses import asdict
from typing import Any, List

from Crypto.Hash import keccak
from eth_typing import ChecksumAddress, HexStr
from hexbytes import HexBytes
from web3 import Web3
from web3.contract.contract import ContractFunction
from web3.types import Wei

from skale.contracts.base_contract import transaction_method
from skale.contracts.manager.node_rotation import NodeRotation
from skale.contracts.manager.schains_internal import SChainsInternal
from skale.contracts.skale_manager_contract import SkaleManagerContract
from skale.types.node import NodeId
from skale.types.schain import (
    SchainHash,
    SchainName,
    SchainStructure,
    SchainStructureWithStatus,
)
from skale.dataclasses.schain_options import (
    SchainOptions,
    get_default_schain_options,
    parse_schain_options,
)


class SChains(SkaleManagerContract):
    """Wrapper for some of the Schains.sol functions"""

    def name_to_group_id(self, name: SchainName) -> HexBytes:
        return self.skale.web3.keccak(text=name)

    @property
    @functools.lru_cache()
    def schains_internal(self) -> SChainsInternal:
        return self.skale.schains_internal

    @property
    @functools.lru_cache()
    def node_rotation(self) -> NodeRotation:
        return self.skale.node_rotation

    def get(self, id_: SchainHash) -> SchainStructure:
        res = self.schains_internal.get_raw(id_)
        options = self.get_options(id_)
        return SchainStructure(**asdict(res), chain_id=self.name_to_id(res.name), options=options)

    def get_by_name(self, name: SchainName) -> SchainStructure:
        id_ = self.name_to_id(name)
        return self.get(id_)

    def get_schains_for_owner(self, account: ChecksumAddress) -> List[SchainStructure]:
        schains = []
        list_size = self.schains_internal.get_schain_list_size(account)

        for i in range(0, list_size):
            id_ = self.schains_internal.get_schain_id_by_index_for_owner(account, i)
            schain = self.get(id_)
            schains.append(schain)
        return schains

    def get_schains_for_node(self, node_id: NodeId) -> list[SchainStructureWithStatus]:
        schains = []
        schain_ids = self.schains_internal.get_schain_ids_for_node(node_id)
        for schain_id in schain_ids:
            simple_schain = self.get(schain_id)
            schain = SchainStructureWithStatus(
                **asdict(simple_schain), active=self.schain_active(simple_schain)
            )
            schains.append(schain)
        return schains

    def get_active_schains_for_node(self, node_id: NodeId) -> List[SchainStructureWithStatus]:
        schains = []
        schain_ids = self.schains_internal.get_active_schain_ids_for_node(node_id)
        for schain_id in schain_ids:
            simple_schain = self.get(schain_id)
            schain = SchainStructureWithStatus(**asdict(simple_schain), active=True)
            schains.append(schain)
        return schains

    def name_to_id(self, name: SchainName) -> SchainHash:
        keccak_hash = keccak.new(data=name.encode('utf8'), digest_bits=256)
        return SchainHash(Web3.to_bytes(hexstr=Web3.to_hex(hexstr=HexStr(keccak_hash.hexdigest()))))

    def get_last_rotation_id(self, schain_name: SchainName) -> int:
        rotation_data = self.node_rotation.get_rotation(schain_name)
        return rotation_data.rotation_counter

    def schain_active(self, schain: SchainStructure) -> bool:
        if (
            schain.name != ''
            and schain.mainnet_owner != '0x0000000000000000000000000000000000000000'
        ):
            return True
        return False

    def get_schain_price(self, index_of_type: int, lifetime: int) -> Wei:
        return Wei(self.contract.functions.getSchainPrice(index_of_type, lifetime).call())

    @transaction_method
    def add_schain_by_foundation(
        self,
        lifetime: int,
        type_of_nodes: int,
        nonce: int,
        name: SchainName,
        options: SchainOptions | None = None,
        schain_owner: ChecksumAddress | None = None,
        schain_originator: ChecksumAddress | None = None,
    ) -> 'ContractFunction':
        if schain_owner is None:
            schain_owner = self.skale.wallet.address
        if schain_originator is None:
            schain_originator = self.skale.wallet.address
        if not options:
            options = get_default_schain_options()

        return self.contract.functions.addSchainByFoundation(
            lifetime,
            type_of_nodes,
            nonce,
            name,
            schain_owner,
            schain_originator,
            options.to_tuples(),
        )

    @transaction_method
    def grant_role(self, role: bytes, owner: ChecksumAddress) -> 'ContractFunction':
        return self.contract.functions.grantRole(role, owner)

    def schain_creator_role(self) -> bytes:
        return bytes(self.contract.functions.SCHAIN_CREATOR_ROLE().call())

    def __raw_get_options(self, schain_id: SchainHash) -> List[Any]:
        return list(self.contract.functions.getOptions(schain_id).call())

    def get_options(self, schain_id: SchainHash) -> SchainOptions:
        return parse_schain_options(raw_options=self.__raw_get_options(schain_id))

    def get_options_by_name(self, name: SchainName) -> SchainOptions:
        id_ = self.name_to_id(name)
        return self.get_options(id_)

    def restart_schain_creation(self, name: SchainName) -> 'ContractFunction':
        return self.contract.functions.restartSchainCreation(name)
